// Mupad-Prozeduren  zur Tomographie - (c) 2004 R. Oldenburg 
// Vereinfachte Version
with(plots): with(LinearAlgebra): with(linalg):

// Grsse des Beckens
Lx:=25: Ly:=39: 
// Auflsung des Bildes  
kx:=2: ky:=2: 
// Grenzen des Bildrechtecks
x1:=5: x2:=15: y1:=15: y2:=25:

initValues:=proc() 
begin 
  n:=kx*ky;  dx:=(x2-x1)/kx; dy:=(y2-y1)/ky;
end_proc:

// Messwerte : Ungeschwcht, Nulleffekt
I0:=358:  nullwert:=3.5: 

// Beispiel 
messwerte:= [
 [8,Lx-8,"na","na",358],
 [13,Lx-13,"na","na",100],
 ["na","na",17,Ly-17,358],
 ["na","na",22,Ly-22,100]
]:
initValues():

// Allgemeine Hilfsprozeduren -------------------------------------------------

Gerade:= proc(p:DOM_LIST,q:DOM_LIST)
  local a,b,c;
  begin
  a:=q[2]-p[2]; b:=p[1]-q[1]; 
  c:=a*p[1]+b*p[2];
  [a,b,c];
end_proc:


Gerade_aus_Eckwerten:=proc(x,xq,y,yq)
  local   S;
  begin
  S:={};
  if testtype(x,Type::Numeric) then S:=S union {[x,0]}; end_if;
  if testtype(xq,Type::Numeric) then S:=S union {[Lx-xq,Ly]}; end_if;
  if testtype(y,Type::Numeric) then S:=S union {[Lx,y]}; end_if;
  if testtype(yq,Type::Numeric) then S:=S union {[0,Ly-yq]}; end_if;
  if nops(S)=2 then Gerade(S[1],S[2]) else print("Fehler",x,xq,y,yq); return([0,0,0]); end_if;
end_proc:


Laenge_in_Zelle:= proc(ix,iy,gg:DOM_LIST)
  local Sly,Sry,Sox,Sux,S,a,b,c,g;
  begin 
  if nops(gg)>3 then g:=Gerade_aus_Eckwerten(gg[1],gg[2],gg[3],gg[4]); 
         	else if nops(gg)=3 then g:=gg; 
				else error("Unzulssige Geraden-Def.".expr2text(g)); end_if; 
  end_if;
  a:=g[1];b:=g[2];c:=g[3];
  if a=0 and b=0 then error("Keine Geradengleichung: a=b=0") end_if;
  if a=0 then  // horizontal
    if c/b>= y1+iy*dy and c/b<y1+(iy+1)*dy then return(float(dx)); else return(0); end_if;
  end_if;
  if b=0 then // vertikal
    if c/a>= x1+ix*dx and c/a<x1+(ix+1)*dx then return(float(dy)); else return(0); end_if;
  end_if;
  // Schnitte mit den Geraden, die die Zelle beranden u_nten, o_ben usw.
  Sly:=c/b-a/b*(x1+ix*dx); Sry:=c/b-a/b*(x1+(ix+1)*dx);
  Sux:=c/a-b/a*(y1+iy*dy); Sox:=c/a-b/a*(y1+(iy+1)*dy);
  S:={};
  if(Sly>=y1+iy*dy and Sly<y1+(iy+1)*dy) then S:=S union {[x1+ix*dx,Sly]}; end_if;
  if(Sry>=y1+iy*dy and Sry<y1+(iy+1)*dy) then S:=S union {[x1+(ix+1)*dx,Sry]}; end_if;
  if(Sux>=x1+ix*dx and Sux<x1+(ix+1)*dx) then S:=S union {[Sux,y1+iy*dy]}; end_if;
  if(Sox>=x1+ix*dx and Sox<x1+(ix+1)*dx) then S:=S union {[Sox,y1+(iy+1)*dy]}; end_if;
  S:=float(S); // Jetzt sollte S die Menge der Schnittpunkt von g mit Zellenstrecken sein.
  if nops(S)<2 then  return(0) ; end_if;
  if nops(S)>3 then print("Seltsames S=",S); end_if; // darf nicht passieren !!!
  if nops(S)=3 then return(max(
    float(sqrt((S[1][1]-S[2][1])^2+(S[1][2]-S[2][2])^2)),
    float(sqrt((S[1][1]-S[3][1])^2+(S[1][2]-S[3][2])^2)),
    float(sqrt((S[3][1]-S[2][1])^2+(S[3][2]-S[2][2])^2)))) 
  end_if;
  float(sqrt((S[1][1]-S[2][1])^2+(S[1][2]-S[2][2])^2));
end_proc:

pos_in_list:=proc(x,l:DOM_LIST) // gibt 0, falls x in l nicht vorkommt, sonst die Position des ersten Auftr.
  local i;
  begin
  for i from 1 to nops(l) do
     if l[i]=x then return(i); end_if;
  end_for;
  return(0);
end_proc:


// Erstellung der Matrizen ------------------------------------------------

make_matrix:= proc(messwerte:DOM_LIST) 
  local gsys,g,i,j,trefferij,s,A,b,vars,delrows,leftout;
  begin
  gsys:={}; vars:=[];
  print(["n,kx,ky",n,kx,ky,"nops(mw)=",nops(messwerte)]);
  delrows:=[]; 
  A:=matrix(nops(messwerte),n,0); b:=matrix(nops(messwerte),1,0);
  for s from 1 to nops(messwerte)  do
   	b[s]:=float(-ln((messwerte[s][5]-nullwert)/I0)); 
	g:=fp::apply(Gerade_aus_Eckwerten,op(messwerte[s][1..4]));
	trefferij:=0;
	for i from 0 to kx-1 do for j from 0 to ky-1 do
		A[s,j*kx+i+1]:=Laenge_in_Zelle(i,j,g);
		if Laenge_in_Zelle(i,j,g)>0 then trefferij:=trefferij+1; end_if;
	end_for; end_for; // Ende Schleife ber Felder
	if trefferij=0 then  delrows:=[op(delrows),s];  end_if;
  end_for;  // Ende der Schleife ber die Strahlen
  A:=matrix(ListDeleteRow(coerce(A,DOM_LIST),delrows)); 
  b:=matrix(ListDeleteRow(coerce(b,DOM_LIST),delrows)); 
  [A,b];
end_proc:

ListDeleteRow:= proc(l,di)
 local i,res;
  begin
 res:=[];
 for i from 1 to nops(l) do
 if not(has(di,i)) then  res:=[op(res),l[i]]; end_if;
 end_for; 
 res
end_proc:

// Verwandeln der in einem Vector gespeicherten mu-Werte in Matrix -----------
Vec2Mat:= proc(v,kx,ky)
   local ix,iy,solF;
   begin
   solF:=matrix(kx,ky):
   for ix from 0 to kx-1 do for iy from 0 to ky-1 do
     solF[ix+1,iy+1]:=v[iy*kx+ix+1]:
   end_for; end_for:
   solF;
 end_proc:

// Vektoren oder Listen aneinaderanhngen
vectorAppend:= proc(a,b)
	local aa,bb;
	begin
		if domtype(a)=DOM_LIST then aa:=op(a) 
		elif domtype(a)=Dom::Matrix() then  aa:=op(map(coerce(a,DOM_LIST),x->x[1]));
		else error("Type error arg 1");
		end_if; 
		if domtype(b)=DOM_LIST then bb:=op(b) 
		elif domtype(b)=Dom::Matrix() then  bb:=op(map(coerce(b,DOM_LIST),x->x[1]));
		else error("Type error arg 1");
		end_if; 
		matrix([aa,bb])
end_proc:

// Mittleres mu fr ganze Probe ------------------------------------------------

muMittel:=proc(messwerte)
  local s,mum,mittelLaenge;
  begin
  mittelLaenge:=float((x2-x1 + y2-y1)/sqrt(2));
  mum:=0;
  for s from 1 to nops(messwerte) do
      mum:= mum+float(-ln((messwerte[s][5]-nullwert)/I0))/mittelLaenge;
  end_for:
  mum/nops(messwerte);
end_proc:

// Rckprojektion --------------------------------------------------------------

BPpropLaenge:=true;

backproj:=proc(messwerte)
 local gs,treffermatrix,laengetot,mu,stot,strahlgerade,i,ix,iy,liz,nz,mustrahl;
 begin
 treffermatrix:=make_gsys(messwerte)[6]: 
 mu:=Matrix(1..kx,1..ky,0):
 for i from 1 to nops(messwerte) do
   stot:=-float(log((messwerte[i][5]-nullwert)/I0));
   strahlgerade:=Gerade_aus_Eckwerten(messwerte[i][1],messwerte[i][2],messwerte[i][3],
	messwerte[i][4]);
   laengetot:=0; // gesamte laenge deieses Strahls im Analyseraster 
   nz:=0; // Zahl der Zellen, die der Strahl trifft
   mustrahl:=Matrix(1..kx,1..ky,0):
   for ix from 0 to kx-1 do for iy from 0 to ky-1 do
     liz:=Laenge_in_Zelle(ix,iy,strahlgerade);
     if liz>0 then nz:=nz+1; end_if;
     laengetot:=laengetot+liz;
     if BPpropLaenge then
        mustrahl[ix+1,iy+1]:=mustrahl[ix+1,iy+1]+liz*stot;
     else
        if liz>0 then mustrahl[ix+1,iy+1]:=mustrahl[ix+1,iy+1]+stot; end_if;
     end_if:    
   end_for: end_for:
   for ix from 1 to kx do for iy from 1 to ky do
   if BPpropLaenge then
     if laengetot=0 then mustrahl[ix,iy]:=0; else  
     mustrahl[ix,iy]:=mustrahl[ix,iy]/laengetot^2;
     end_if;
   else
     if nz=0 then mustrahl[ix,iy]:= 0 
     else  mustrahl[ix,iy]:= mustrahl[ix,iy]/nz/laengetot; end_if;
   end_if:  
   end_for; end_for;
   mu:=mu+mustrahl;
 end_for:
 for ix from 1 to kx do for iy from 1 to ky do
   mu[ix,iy]:=mu[ix,iy]/treffermatrix[ix,iy];
 end_for: end_for:
 mu
end_proc:

make_gsys:= proc(messwerte:DOM_LIST) 
 local A0,b0,dull,gsys,ab,g,i,j,s,A,b,vars,delrows,leftout,treffer;
 begin
 gsys:={}; vars:=[];
 print(["n,kx,ky",n,kx,ky,"nops(mw)=",nops(messwerte)]);
 treffer:=matrix(kx,ky,0); delrows:=[]; 
 A:=matrix(nops(messwerte),1..n,0); b:=matrix(1..nops(messwerte),1,0);
 for s from 1 to nops(messwerte)  do
 b[s]:=float(-log((messwerte[s][5]-nullwert)/I0)); 
 ab:=0;
 g:=apply(Gerade_aus_Eckwerten,op(messwerte[s][1..4]));
 for i from 0 to kx-1 do for j from 0 to ky-1 do
 A[s,j*kx+i+1]:=Laenge_in_Zelle(i,j,g); 
 if false then print("Strahl ",s,"Gerade",g,
   "Zelle ",[i,j],"pos",[x1+i*dx,y1+j*dy],"Laenge=",-A[s,j*kx+i+1]);
 end_if; 
 ab:=ab+A[s,j*kx+i+1]*c[i,j]; 
 if A[s,j*kx+i+1]>0 then treffer[i+1,j+1]:=treffer[i+1,j+1]+1; end_if;
 if s=1 then vars:=[op(vars),c[i,j]]; end_if;
 end_for; end_for; // Ende Schleife ber Felder
 if not(ab=0) then gsys:=gsys union {b[s]=ab}; 
 else delrows:=[op(delrows),s];  end_if;
 end_for;  // Ende der Schleife ber die Strahlen
 A0:=A; b0:=b;
 print("Eliminating rows ",delrows);
 A:=DeleteRow(A,delrows); b:=convert(DeleteRow(convert(b,Matrix),delrows),Vector); 
 dull:={op(vars)} minus indets(gsys);
 if nops(dull)>0 then print("Undetermined vars:",dull); end_if;
 leftout:=map(v->pos_in_list(v,vars),[op(dull)]);
 A:=DeleteColumn(A,leftout);
 [gsys,vars,A,b,leftout,treffer,A0,b0];
end_proc:

 
// Graphische tools ------------------------------------------------------------ 

displayrays:= proc(messwerte)
  begin
  showray(map( messwerte,L -> Gerade_aus_Eckwerten(L[1],L[2],L[3],L[4])));
end_proc:

showray:= proc(gL:DOM_LIST)
 local becken,feld,gitter,strahlen,LinieSchw,i;
  begin
    LinieSchw:= coords -> PLOT(CURVES(coords,COLOR(RGB,0,0,0))):
 
    becken:=plot::Polygon2d([[0,0],[Lx,0],[Lx,Ly],[0,Ly],[0,0]]);
    feld:=plot::Polygon2d([[x1,y1],[x1,y2],[x2,y2],[x2,y1],[x1,y1]],
	     Color=RGB::Green);
    gitter:=[plot::Polygon2d([[x1+i*dx,y1],[x1+i*dx,y2]],Color=RGB::Blue)$i=0..kx,
             plot::Polygon2d([[x1,y1+i*dy],[x2,y1+i*dy]],Color=RGB::Blue)$i=0..ky];
    strahlen:=map( gL,g ->  plot::Implicit2d(g[1]*x+g[2]*y=g[3],x=0..Lx,y=0..Ly,Color=RGB::Red) );
    plot(becken,op(gitter),feld,
          op(strahlen),
   Axes=Boxed,Scaling=Constrained  /* ,view=[-1..Lx+1,-1..Ly+1]  */ );
 end_proc:




PlotMatrix:= A -> plot(plot::Density(A,FillColor2=RGB::Black,FillColor=RGB::White)):

:
